# -*- mode: python; coding: utf-8 -*-

# Copyright © 2017 by Jeffrey C. Ollie <jeff@ocjtech.us>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

import argparse
import sys
import yaml
import re

from ldaptor.protocols.ldap.ldapclient import LDAPClient
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry

from twisted.internet import reactor
from twisted.internet.endpoints import clientFromString
from twisted.internet.endpoints import serverFromString
from twisted.internet.protocol import Factory
from twisted.logger import Logger
from twisted.logger import globalLogBeginner
from twisted.logger import textFileLogObserver
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.web.server import Site

class QuietSite(Site):
   noisy = False

class LDAPFactory(Factory):
   noisy = False

   def buildProtocol(self, address):
      return LDAPClient()

class Metrics(object):
   log = Logger()

   basedn = 'cn=Monitor'
   basedn_replication = 'cn=config'
   query = '(cn=monitor)'
   query_replication = '(objectClass=nsds5replicationagreement)'

   def __init__(self, request, config):
      self.request = request
      self.config = config
      factory = LDAPFactory()
      endpoint = clientFromString(reactor, self.config['client'])
      d = endpoint.connect(factory)
      d.addCallback(self.gotConnection)

   def gotConnection(self, client):
      self.client = client
      if 'binddn' in self.config:
         d = self.client.bind(self.config['binddn'], self.config['bindpw'])
      else:
         d = self.client.bind()
      d.addCallback(self.isAuthenticated)

   def isAuthenticated(self, result):
      base = LDAPEntry(self.client, self.basedn)
      d = base.search(filterText = self.query, attributes = ('*', '+'))

      base_replication = LDAPEntry(self.client, self.basedn_replication)
      d_replication = base_replication.search(filterText = self.query_replication, attributes = ('*', '+'))

      d.addCallback(self.gotResults)
      d_replication.addCallback(self.gotResultsReplication)

   def gotResults(self, results):
      self.request.setHeader(b'Content-Type', b'text/plain; charset=utf-8; version=0.0.4')

      self.request.write('dirsrv_up 1\n'.encode('utf-8'))

      for entry in results:
         for metric in self.config['metrics']:
             try:
                value = float(entry[metric].pop())
                self.request.write('dirsrv_monitor_' + metric + ' {}\n'.format(value).encode('utf-8'))
             except (KeyError, ValueError):
                pass

   def gotResultsReplication(self, results_replicaton):
      for entry in results_replicaton:
         try:
            name = 'peer="{}"'.format(entry['nsDS5ReplicaHost'].pop())
            root = 'root="{}"'.format(entry['nsDS5ReplicaRoot'].pop())
            raw_value = entry['nsds5replicaLastUpdateStatus'].pop()
            value = int(re.findall(r'\d+',raw_value)[0])
            self.request.write('dirsrv_replication_agreement_status{{{},{}}} {}\n'.format(name, root, value).encode('utf-8'))
         except ValueError:
            pass

      self.request.finish()
      self.client.unbind()
      self.client.transport.loseConnection()

class MetricsPage(Resource):
   log = Logger()
   isLeaf = True

   def __init__(self, config):
      self.config = config
      Resource.__init__(self)

   def render_GET(self, request):
      Metrics(request, config)
      return NOT_DONE_YET

class RootPage(Resource):
   isLeaf = False

   def render_GET(self, request):
      request.setHeader(b'Content-Type', b'text/plain; charset=utf-8')
      return 'OK\n'.encode('utf-8')

parser = argparse.ArgumentParser(prog = '389_exporter',
                                 description = 'Prometheus 389 Directory Server exporter')
parser.add_argument('--config',
                    type = argparse.FileType('r'),
                    help = 'configuration file',
                    required = True)
arguments = parser.parse_args()

config = yaml.load(arguments.config)
arguments.config.close()

output = textFileLogObserver(sys.stderr, timeFormat='')
globalLogBeginner.beginLoggingTo([output])

metrics = MetricsPage(config)
root = RootPage()
root.putChild(b'metrics', metrics)
site = QuietSite(root)
endpoint = serverFromString(reactor, config['server'])
endpoint.listen(site)

reactor.run()
